@@ -1,5 +1,5 @@
 defmodule IO do
-  @moduledoc """
+  @moduledoc ~S"""
   Functions handling input/output (IO).
 
   Many functions in this module expect an IO device as an argument.
@@ -7,13 +7,10 @@ defmodule IO do
   For convenience, Elixir provides `:stdio` and `:stderr` as
   shortcuts to Erlang's `:standard_io` and `:standard_error`.
 
-  The majority of the functions expect chardata, i.e. strings or
-  lists of characters and strings. In case another type is given,
-  functions will convert to string via the `String.Chars` protocol
-  (as shown in typespecs).
-
-  The functions starting with `bin` expect iodata as an argument,
-  i.e. binaries or lists of bytes and binaries.
+  The majority of the functions expect chardata. In case another type is given,
+  functions will convert those types to string via the `String.Chars` protocol
+  (as shown in typespecs). For more information on chardata, see the
+  "IO data" section below.
 
   ## IO devices
 
@@ -27,22 +24,105 @@ defmodule IO do
     * `:stderr` - a shortcut for the named process `:standard_error`
       provided in Erlang
 
-  IO devices maintain their position, that means subsequent calls to any
-  reading or writing functions will start from the place when the device
-  was last accessed. Position of files can be changed using the
+  IO devices maintain their position, which means subsequent calls to any
+  reading or writing functions will start from the place where the device
+  was last accessed. The position of files can be changed using the
   `:file.position/2` function.
 
+  ## IO data
+
+  IO data is a data type that can be used as a more efficient alternative to binaries
+  in certain situations.
+
+  A term of type **IO data** is a binary or a list containing bytes (integers in `0..255`)
+  or nested IO data. The type is recursive. Let's see an example of one of
+  the possible IO data representing the binary `"hello"`:
+
+      [?h, "el", ["l", [?o]]]
+
+  The built-in `t:iodata/0` type is defined in terms of `t:iolist/0`. An IO list is
+  the same as IO data but it doesn't allow for a binary at the top level (but binaries
+  are still allowed in the list itself).
+
+  ### Use cases for IO data
+
+  IO data exists because often you need to do many append operations
+  on smaller chunks of binaries in order to create a bigger binary. However, in
+  Erlang and Elixir concatenating binaries will copy the concatenated binaries
+  into a new binary.
+
+      def email(username, domain) do
+        username <> "@" <> domain
+      end
+
+  In this function, creating the email address will copy the `username` and `domain`
+  binaries. Now imagine you want to use the resulting email inside another binary:
+
+      def welcome_message(name, username, domain) do
+        "Welcome #{name}, your email is: #{email(username, domain)}"
+      end
+
+      IO.puts(welcome_message("Meg", "meg", "example.com"))
+      #=> "Welcome Meg, your email is: meg@example.com"
+
+  Every time you concatenate binaries or use interpolation (`#{}`) you are making
+  copies of those binaries. However, in many cases you don't need the complete
+  binary while you create it, but only at the end to print it out or send it
+  somewhere. In such cases, you can construct the binary by creating IO data:
+
+      def email(username, domain) do
+        [username, ?@, domain]
+      end
+
+      def welcome_message(name, username, domain) do
+        ["Welcome ", name, ", your email is: ", email(username, domain)]
+      end
+
+      IO.puts(welcome_message("Meg", "meg", "example.com"))
+      #=> "Welcome Meg, your email is: meg@example.com"
+
+  Building IO data is cheaper than concatenating binaries. Concatenating multiple
+  pieces of IO data just means putting them together inside a list since IO data
+  can be arbitrarily nested, and that's a cheap and efficient operation. Most of
+  the IO-based APIs, such as `:gen_tcp` and `IO`, receive IO data and write it
+  to the socket directly without converting it to binary.
+
+  One drawback of IO data is that you can't do things like pattern match on the
+  first part of a piece of IO data like you can with a binary, because you usually
+  don't know the shape of the IO data. In those cases, you may need to convert it
+  to a binary by calling `iodata_to_binary/1`, which is reasonably efficient
+  since it's implemented natively in C. Other functionality, like computing the
+  length of IO data, can be computed directly on the iodata by calling `iodata_length/1`.
+
+  ### Chardata
+
+  Erlang and Elixir also have the idea of `t:chardata/0`. Chardata is very
+  similar to IO data: the only difference is that integers in IO data represent
+  bytes while integers in chardata represent Unicode code points. Bytes
+  (`t:byte/0`) are integers in the `0..255` range, while Unicode code points
+  (`t:char/0`) are integers in the range `0..0x10FFFF`. The `IO` module provides
+  the `chardata_to_string/1` function for chardata as the "counter-part" of the
+  `iodata_to_binary/1` function for IO data.
+
+  If you try to use `iodata_to_binary/1` on chardata, it will result in an
+  argument error. For example, let's try to put a code point that is not
+  representable with one byte, like `?π`, inside IO data:
+
+      iex> IO.iodata_to_binary(["The symbol for pi is: ", ?π])
+      ** (ArgumentError) argument error
+
+  If we use chardata instead, it will work as expected:
+
+      iex> IO.chardata_to_string(["The symbol for pi is: ", ?π])
+      "The symbol for pi is: π"
+
   """
 
   @type device :: atom | pid
   @type nodata :: {:error, term} | :eof
-  @type chardata() :: :unicode.chardata()
+  @type chardata :: String.t() | maybe_improper_list(char | chardata, String.t() | [])
 
-  defmacrop is_iodata(data) do
-    quote do
-      is_list(unquote(data)) or is_binary(unquote(data))
-    end
-  end
+  defguardp is_iodata(data) when is_list(data) or is_binary(data)
 
   @doc """
   Reads from the IO `device`.
@@ -68,7 +148,7 @@ defmodule IO do
   def read(device \\ :stdio, line_or_chars)
 
   def read(device, :all) do
-    do_read_all(map_dev(device), "")
+    do_read_all(map_dev(device), :empty)
   end
 
   def read(device, :line) do
@@ -81,12 +161,27 @@ defmodule IO do
 
   defp do_read_all(mapped_dev, acc) do
     case :io.get_line(mapped_dev, "") do
-      line when is_binary(line) -> do_read_all(mapped_dev, acc <> line)
-      :eof -> acc
+      line when is_binary(line) or is_list(line) -> do_read_all(mapped_dev, concat(acc, line))
+      :eof -> read_eof(mapped_dev, acc)
       other -> other
     end
   end
 
+  defp concat(:empty, line), do: line
+  defp concat(acc, line) when is_binary(acc), do: acc <> line
+  defp concat(acc, line) when is_list(acc), do: acc ++ line
+
+  defp read_eof(device, :empty) do
+    with [_ | _] = opts <- :io.getopts(device),
+         false <- Keyword.get(opts, :binary, true) do
+      ''
+    else
+      _ -> ""
+    end
+  end
+
+  defp read_eof(_device, acc), do: acc
+
   @doc """
   Reads from the IO `device`. The operation is Unicode unsafe.
 
@@ -141,47 +236,65 @@ defmodule IO do
   end
 
   @doc """
-  Writes `item` to the given `device`.
+  Writes `chardata` to the given `device`.
 
-  By default the `device` is the standard output.
-  It returns `:ok` if it succeeds.
+  By default, the `device` is the standard output.
 
   ## Examples
 
-      IO.write "sample"
+      IO.write("sample")
       #=> sample
 
-      IO.write :stderr, "error"
+      IO.write(:stderr, "error")
       #=> error
 
   """
-  @spec write(device, chardata | String.Chars.t) :: :ok
-  def write(device \\ :stdio, item) do
-    :io.put_chars map_dev(device), to_chardata(item)
+  @spec write(device, chardata | String.Chars.t()) :: :ok
+  def write(device \\ :stdio, chardata) do
+    :io.put_chars(map_dev(device), to_chardata(chardata))
   end
 
   @doc """
-  Writes `item` as a binary to the given `device`.
-  No Unicode conversion happens.
-  The operation is Unicode unsafe.
+  Writes `iodata` to the given `device`.
 
-  Check `write/2` for more information.
+  This operation is meant to be used with "raw" devices
+  that are started without an encoding. The given `iodata`
+  is written as is to the device, without conversion. For
+  more information on IO data, see the "IO data" section in
+  the module documentation.
 
-  Note: do not use this function on IO devices in Unicode mode
-  as it will return the wrong result.
+  Use `write/2` for devices with encoding.
+
+  Important: do **not** use this function on IO devices in
+  Unicode mode as it will write the wrong data. In particular,
+  the standard IO device is set to Unicode by default, so writing
+  to stdio with this function will likely result in the wrong data
+  being sent down the wire.
   """
   @spec binwrite(device, iodata) :: :ok | {:error, term}
-  def binwrite(device \\ :stdio, item) when is_iodata(item) do
-    :file.write map_dev(device), item
+  def binwrite(device \\ :stdio, iodata) when is_iodata(iodata) do
+    :file.write(map_dev(device), iodata)
   end
 
   @doc """
   Writes `item` to the given `device`, similar to `write/2`,
   but adds a newline at the end.
+
+  By default, the `device` is the standard output. It returns `:ok`
+  if it succeeds.
+
+  ## Examples
+
+      IO.puts("Hello World!")
+      #=> Hello World!
+
+      IO.puts(:stderr, "error")
+      #=> error
+
   """
-  @spec puts(device, chardata | String.Chars.t) :: :ok
+  @spec puts(device, chardata | String.Chars.t()) :: :ok
   def puts(device \\ :stdio, item) do
-    :io.put_chars map_dev(device), [to_chardata(item), ?\n]
+    :io.put_chars(map_dev(device), [to_chardata(item), ?\n])
   end
 
   @doc """
@@ -196,18 +309,41 @@ defmodule IO do
   ## Examples
 
       stacktrace = [{MyApp, :main, 1, [file: 'my_app.ex', line: 4]}]
-      IO.warn "variable bar is unused", stacktrace
+      IO.warn("variable bar is unused", stacktrace)
       #=> warning: variable bar is unused
       #=>   my_app.ex:4: MyApp.main/1
 
   """
-  @spec warn(chardata | String.Chars.t, Exception.stacktrace) :: :ok
+  @spec warn(chardata | String.Chars.t(), Exception.stacktrace()) :: :ok
   def warn(message, []) do
-    :elixir_errors.warn([to_chardata(message), ?\n])
+    message = [to_chardata(message), ?\n]
+    :elixir_errors.io_warn(0, nil, message, message)
   end
-  def warn(message, stacktrace) when is_list(stacktrace) do
-    formatted = Enum.map_join(stacktrace, "\n  ", &Exception.format_stacktrace_entry(&1))
-    :elixir_errors.warn([to_chardata(message), ?\n, "  ", formatted, ?\n])
+
+  def warn(message, [{_, _, _, opts} | _] = stacktrace) do
+    message = to_chardata(message)
+    formatted_trace = Enum.map_join(stacktrace, "\n  ", &Exception.format_stacktrace_entry(&1))
+    line = opts[:line]
+    file = opts[:file]
+
+    :elixir_errors.io_warn(
+      line || 0,
+      file && List.to_string(file),
+      message,
+      [message, ?\n, "  ", formatted_trace, ?\n]
+    )
+  end
+
+  @doc false
+  def warn_once(key, message, stacktrace_drop_levels) do
+    {:current_stacktrace, stacktrace} = Process.info(self(), :current_stacktrace)
+    stacktrace = Enum.drop(stacktrace, stacktrace_drop_levels)
+
+    if :elixir_config.warn(key, stacktrace) do
+      warn(message, stacktrace)
+    else
+      :ok
+    end
   end
 
   @doc """
@@ -215,14 +351,19 @@ defmodule IO do
 
   It returns `:ok` if it succeeds.
 
+  Do not call this function at the tail of another function. Due to tail
+  call optimization, a stacktrace entry would not be added and the
+  stacktrace would be incorrectly trimmed. Therefore make sure at least
+  one expression (or an atom such as `:ok`) follows the `IO.warn/1` call.
+
   ## Examples
 
-      IO.warn "variable bar is unused"
+      IO.warn("variable bar is unused")
       #=> warning: variable bar is unused
       #=>   (iex) evaluator.ex:108: IEx.Evaluator.eval/4
 
   """
-  @spec warn(chardata | String.Chars.t) :: :ok
+  @spec warn(chardata | String.Chars.t()) :: :ok
   def warn(message) do
     {:current_stacktrace, stacktrace} = Process.info(self(), :current_stacktrace)
     warn(message, Enum.drop(stacktrace, 2))
@@ -248,7 +389,7 @@ defmodule IO do
 
   ## Examples
 
-      IO.inspect <<0, 1, 2>>, width: 40
+      IO.inspect(<<0, 1, 2>>, width: 40)
 
   Prints:
 
@@ -256,7 +397,7 @@ defmodule IO do
 
   We can use the `:label` option to decorate the output:
 
-      IO.inspect 1..100, label: "a wonderful range"
+      IO.inspect(1..100, label: "a wonderful range")
 
   Prints:
 
@@ -268,7 +409,7 @@ defmodule IO do
       |> IO.inspect(label: "before")
       |> Enum.map(&(&1 * 2))
       |> IO.inspect(label: "after")
-      |> Enum.sum
+      |> Enum.sum()
 
   Prints:
 
@@ -276,9 +417,9 @@ defmodule IO do
       after: [2, 4, 6]
 
   """
-  @spec inspect(item, Keyword.t) :: item when item: var
+  @spec inspect(item, keyword) :: item when item: var
   def inspect(item, opts \\ []) do
-    inspect :stdio, item, opts
+    inspect(:stdio, item, opts)
   end
 
   @doc """
@@ -286,12 +427,13 @@ defmodule IO do
 
   See `inspect/2` for a full list of options.
   """
-  @spec inspect(device, item, Keyword.t) :: item when item: var
+  @spec inspect(device, item, keyword) :: item when item: var
   def inspect(device, item, opts) when is_list(opts) do
-    label    = if (label = opts[:label]), do: [to_chardata(label), ": "], else: []
-    opts     = struct(Inspect.Opts, opts)
-    chardata = Inspect.Algebra.format(Inspect.Algebra.to_doc(item, opts), opts.width)
-    puts device, [label, chardata]
+    label = if label = opts[:label], do: [to_chardata(label), ": "], else: []
+    opts = struct(Inspect.Opts, opts)
+    doc = Inspect.Algebra.group(Inspect.Algebra.to_doc(item, opts))
+    chardata = Inspect.Algebra.format(doc, opts.width)
+    puts(device, [label, chardata])
     item
   end
 
@@ -299,14 +441,14 @@ defmodule IO do
   Gets a number of bytes from IO device `:stdio`.
 
   If `:stdio` is a Unicode device, `count` implies
-  the number of Unicode codepoints to be retrieved.
+  the number of Unicode code points to be retrieved.
   Otherwise, `count` is the number of raw bytes to be retrieved.
 
   See `IO.getn/3` for a description of return values.
 
   """
-  @spec getn(chardata | String.Chars.t, pos_integer) :: chardata | nodata
-  @spec getn(device, chardata | String.Chars.t) :: chardata | nodata
+  @spec getn(device | chardata | String.Chars.t(), pos_integer | chardata | String.Chars.t()) ::
+          chardata | nodata
   def getn(prompt, count \\ 1)
 
   def getn(prompt, count) when is_integer(count) and count > 0 do
@@ -321,7 +463,7 @@ defmodule IO do
   Gets a number of bytes from the IO `device`.
 
   If the IO `device` is a Unicode device, `count` implies
-  the number of Unicode codepoints to be retrieved.
+  the number of Unicode code points to be retrieved.
   Otherwise, `count` is the number of raw bytes to be retrieved.
 
   It returns:
@@ -335,7 +477,7 @@ defmodule IO do
       NFS volume
 
   """
-  @spec getn(device, chardata | String.Chars.t, pos_integer) :: chardata | nodata
+  @spec getn(device, chardata | String.Chars.t(), pos_integer) :: chardata | nodata
   def getn(device, prompt, count) when is_integer(count) and count > 0 do
     :io.get_chars(map_dev(device), to_chardata(prompt), count)
   end
@@ -358,10 +500,10 @@ defmodule IO do
 
   To display "What is your name?" as a prompt and await user input:
 
-      IO.gets "What is your name?\n"
+      IO.gets("What is your name?\n")
 
   """
-  @spec gets(device, chardata | String.Chars.t) :: chardata | nodata
+  @spec gets(device, chardata | String.Chars.t()) :: chardata | nodata
   def gets(device \\ :stdio, prompt) do
     :io.get_line(map_dev(device), to_chardata(prompt))
   end
@@ -387,10 +529,10 @@ defmodule IO do
   Here is an example on how we mimic an echo server
   from the command line:
 
-      Enum.each IO.stream(:stdio, :line), &IO.write(&1)
+      Enum.each(IO.stream(:stdio, :line), &IO.write(&1))
 
   """
-  @spec stream(device, :line | pos_integer) :: Enumerable.t
+  @spec stream(device, :line | pos_integer) :: Enumerable.t()
   def stream(device, line_or_codepoints)
       when line_or_codepoints == :line
       when is_integer(line_or_codepoints) and line_or_codepoints > 0 do
@@ -415,7 +557,7 @@ defmodule IO do
   mode as it will return the wrong result.
 
   """
-  @spec binstream(device, :line | pos_integer) :: Enumerable.t
+  @spec binstream(device, :line | pos_integer) :: Enumerable.t()
   def binstream(device, line_or_bytes)
       when line_or_bytes == :line
       when is_integer(line_or_bytes) and line_or_bytes > 0 do
@@ -423,8 +565,10 @@ defmodule IO do
   end
 
   @doc """
-  Converts chardata (a list of integers representing codepoints,
-  lists and strings) into a string.
+  Converts chardata into a string.
+
+  For more information about chardata, see the ["Chardata"](#module-chardata)
+  section in the module documentation.
 
   In case the conversion fails, it raises an `UnicodeConversionError`.
   If a string is given, it returns the string itself.
@@ -441,7 +585,7 @@ defmodule IO do
       "string"
 
   """
-  @spec chardata_to_string(chardata) :: String.t | no_return
+  @spec chardata_to_string(chardata) :: String.t()
   def chardata_to_string(string) when is_binary(string) do
     string
   end
@@ -451,14 +595,16 @@ defmodule IO do
   end
 
   @doc """
-  Converts iodata (a list of integers representing bytes, lists
-  and binaries) into a binary.
+  Converts IO data into a binary
+
   The operation is Unicode unsafe.
 
-  Notice that this function treats lists of integers as raw bytes
-  and does not perform any kind of encoding conversion. If you want
-  to convert from a charlist to a string (UTF-8 encoded), please
-  use `chardata_to_string/1` instead.
+  Note that this function treats integers in the given IO data as
+  raw bytes and does not perform any kind of encoding conversion.
+  If you want to convert from a charlist to a UTF-8-encoded string,
+  use `chardata_to_string/1` instead. For more information about
+  IO data and chardata, see the ["IO data"](#module-io-data) section in the
+  module documentation.
 
   If this function receives a binary, the same binary is returned.
 
@@ -478,12 +624,15 @@ defmodule IO do
 
   """
   @spec iodata_to_binary(iodata) :: binary
-  def iodata_to_binary(item) do
-    :erlang.iolist_to_binary(item)
+  def iodata_to_binary(iodata) do
+    :erlang.iolist_to_binary(iodata)
   end
 
   @doc """
-  Returns the size of an iodata.
+  Returns the size of an IO data.
+
+  For more information about IO data, see the ["IO data"](#module-io-data)
+  section in the module documentation.
 
   Inlined by the compiler.
 
@@ -494,8 +643,8 @@ defmodule IO do
 
   """
   @spec iodata_length(iodata) :: non_neg_integer
-  def iodata_length(item) do
-    :erlang.iolist_size(item)
+  def iodata_length(iodata) do
+    :erlang.iolist_size(iodata)
   end
 
   @doc false
@@ -503,8 +652,10 @@ defmodule IO do
     case read(device, line_or_codepoints) do
       :eof ->
         {:halt, device}
+
       {:error, reason} ->
         raise IO.StreamError, reason: reason
+
       data ->
         {[data], device}
     end
@@ -515,8 +666,10 @@ defmodule IO do
     case binread(device, line_or_chars) do
       :eof ->
         {:halt, device}
+
       {:error, reason} ->
         raise IO.StreamError, reason: reason
+
       data ->
         {[data], device}
     end
@@ -525,7 +678,7 @@ defmodule IO do
   @compile {:inline, map_dev: 1, to_chardata: 1}
 
   # Map the Elixir names for standard IO and error to Erlang names
-  defp map_dev(:stdio),  do: :standard_io
+  defp map_dev(:stdio), do: :standard_io
   defp map_dev(:stderr), do: :standard_error
   defp map_dev(other) when is_atom(other) or is_pid(other) or is_tuple(other), do: other
 
