@@ -1,6 +1,6 @@
 defmodule List do
   @moduledoc """
-  Functions that work on (linked) lists.
+  Linked lists hold zero, one, or more elements in the chosen order.
 
   Lists in Elixir are specified between square brackets:
 
@@ -45,65 +45,89 @@ defmodule List do
   slower as the list grows in size (linear time):
 
       iex> list = [1, 2, 3]
-      iex> [0 | list]   # fast
+      iex> [0 | list] # fast
       [0, 1, 2, 3]
-      iex> list ++ [4]  # slow
+      iex> list ++ [4] # slow
       [1, 2, 3, 4]
 
-  The `Kernel` module contains many functions to manipulate lists
-  and that are allowed in guards. For example, `Kernel.hd/1` to
-  retrieve the head, `Kernel.tl/1` to fetch the tail and
-  `Kernel.length/1` for calculating the length. Keep in mind that,
-  similar to appending to a list, calculating the length needs to
-  traverse the whole list.
+  Most of the functions in this module work in linear time. This means that,
+  that the time it takes to perform an operation grows at the same rate as the
+  length of the list. For example `length/1` and `last/1` will run in linear
+  time because they need to iterate through every element of the list, but
+  `first/1` will run in constant time because it only needs the first element.
+
+  Lists also implement the `Enumerable` protocol, so many functions to work with
+  lists are found in the `Enum` module. Additionally, the following functions and
+  operators for lists are found in `Kernel`:
+
+    * `++/2`
+    * `--/2`
+    * `hd/1`
+    * `tl/1`
+    * `in/2`
+    * `length/1`
 
   ## Charlists
 
-  If a list is made of non-negative integers, it can also be called
-  a charlist. Elixir uses single quotes to define charlists:
+  If a list is made of non-negative integers, where each integer represents a
+  Unicode code point, the list can also be called a charlist. These integers
+  must:
+
+    * be within the range `0..0x10FFFF` (`0..1_114_111`);
+    * and be out of the range `0xD800..0xDFFF` (`55_296..57_343`), which is
+      reserved in Unicode for UTF-16 surrogate pairs.
+
+  Elixir uses single quotes to define charlists:
 
       iex> 'héllo'
       [104, 233, 108, 108, 111]
 
-  In particular, charlists may be printed back in single
-  quotes if they contain only ASCII-printable codepoints:
+  In particular, charlists will be printed back by default in single
+  quotes if they contain only printable ASCII characters:
 
       iex> 'abc'
       'abc'
 
+  Even though the representation changed, the raw data does remain a list of
+  numbers, which can be handled as such:
+
+      iex> inspect('abc', charlists: :as_list)
+      "[97, 98, 99]"
+      iex> Enum.map('abc', fn num -> 1000 + num end)
+      [1097, 1098, 1099]
+
+  You can use the `IEx.Helpers.i/1` helper to get a condensed rundown on
+  charlists in IEx when you encounter them, which shows you the type, description
+  and also the raw representation in one single summary.
+
   The rationale behind this behaviour is to better support
   Erlang libraries which may return text as charlists
-  instead of Elixir strings. One example of such functions
-  is `Application.loaded_applications/0`:
+  instead of Elixir strings. In Erlang, charlists are the default
+  way of handling strings, while in Elixir it's binaries. One
+  example of such functions is `Application.loaded_applications/0`:
 
-      Application.loaded_applications
-      #=>  [{:stdlib, 'ERTS  CXC 138 10', '2.6'},
-            {:compiler, 'ERTS  CXC 138 10', '6.0.1'},
-            {:elixir, 'elixir', '1.0.0'},
-            {:kernel, 'ERTS  CXC 138 10', '4.1'},
-            {:logger, 'logger', '1.0.0'}]
+      Application.loaded_applications()
+      #=>  [
+      #=>    {:stdlib, 'ERTS  CXC 138 10', '2.6'},
+      #=>    {:compiler, 'ERTS  CXC 138 10', '6.0.1'},
+      #=>    {:elixir, 'elixir', '1.0.0'},
+      #=>    {:kernel, 'ERTS  CXC 138 10', '4.1'},
+      #=>    {:logger, 'logger', '1.0.0'}
+      #=>  ]
 
-  ## List and Enum modules
+  A list can be checked if it is made of only printable ASCII
+  characters with `ascii_printable?/2`.
 
-  This module aims to provide operations that are specific
-  to lists, like conversion between data types, updates,
-  deletions and key lookups (for lists of tuples). For traversing
-  lists in general, developers should use the functions in the
-  `Enum` module that work across a variety of data types.
-
-  In both `Enum` and `List` modules, any kind of index access
-  on a list is linear. Negative indexes are also supported but
-  they imply the list will be iterated twice, one to calculate
-  the proper index and another to perform the operation.
+  Improper lists are never deemed as charlists.
   """
 
   @compile :inline_list_funcs
 
   @doc """
-  Deletes the given `item` from the `list`. Returns a new list without
-  the item.
+  Deletes the given `element` from the `list`. Returns a new list without
+  the element.
 
-  If the `item` occurs more than once in the `list`, just
+  If the `element` occurs more than once in the `list`, just
   the first occurrence is removed.
 
   ## Examples
@@ -111,28 +135,47 @@ defmodule List do
       iex> List.delete([:a, :b, :c], :a)
       [:b, :c]
 
+      iex> List.delete([:a, :b, :c], :d)
+      [:a, :b, :c]
+
       iex> List.delete([:a, :b, :b, :c], :b)
       [:a, :b, :c]
 
+      iex> List.delete([], :b)
+      []
+
   """
-  @spec delete(list, any) :: list
-  def delete(list, item) do
-    :lists.delete(item, list)
-  end
+  @spec delete([], any) :: []
+  @spec delete([...], any) :: list
+  def delete(list, element)
+  def delete([element | list], element), do: list
+  def delete([other | list], element), do: [other | delete(list, element)]
+  def delete([], _element), do: []
 
   @doc """
   Duplicates the given element `n` times in a list.
 
+  `n` is an integer greater than or equal to `0`.
+
+  If `n` is `0`, an empty list is returned.
+
   ## Examples
 
-      iex> List.duplicate("hello", 3)
-      ["hello", "hello", "hello"]
+      iex> List.duplicate("hello", 0)
+      []
 
-      iex> List.duplicate([1, 2], 2)
-      [[1, 2], [1, 2]]
+      iex> List.duplicate("hi", 1)
+      ["hi"]
+
+      iex> List.duplicate("bye", 2)
+      ["bye", "bye"]
+
+      iex> List.duplicate([1, 2], 3)
+      [[1, 2], [1, 2], [1, 2]]
 
   """
-  @spec duplicate(elem, non_neg_integer) :: [elem] when elem: var
+  @spec duplicate(any, 0) :: []
+  @spec duplicate(elem, pos_integer) :: [elem, ...] when elem: var
   def duplicate(elem, n) do
     :lists.duplicate(n, elem)
   end
@@ -140,11 +183,16 @@ defmodule List do
   @doc """
   Flattens the given `list` of nested lists.
 
+  Empty list elements are discarded.
+
   ## Examples
 
       iex> List.flatten([1, [[2], 3]])
       [1, 2, 3]
 
+      iex> List.flatten([[], [[], []]])
+      []
+
   """
   @spec flatten(deep_list) :: list when deep_list: [any | deep_list]
   def flatten(list) do
@@ -156,11 +204,17 @@ defmodule List do
   The list `tail` will be added at the end of
   the flattened list.
 
+  Empty list elements from `list` are discarded,
+  but not the ones from `tail`.
+
   ## Examples
 
       iex> List.flatten([1, [[2], 3]], [4, 5])
       [1, 2, 3, 4, 5]
 
+      iex> List.flatten([1, [], 2], [3, [], 4])
+      [1, 2, 3, [], 4]
+
   """
   @spec flatten(deep_list, [elem]) :: [elem] when elem: var, deep_list: [elem | deep_list]
   def flatten(list, tail) do
@@ -173,16 +227,16 @@ defmodule List do
 
   ## Examples
 
-      iex> List.foldl([5, 5], 10, fn(x, acc) -> x + acc end)
+      iex> List.foldl([5, 5], 10, fn x, acc -> x + acc end)
       20
 
-      iex> List.foldl([1, 2, 3, 4], 0, fn(x, acc) -> x - acc end)
+      iex> List.foldl([1, 2, 3, 4], 0, fn x, acc -> x - acc end)
       2
 
   """
   @spec foldl([elem], acc, (elem, acc -> acc)) :: acc when elem: var, acc: var
-  def foldl(list, acc, function) when is_list(list) and is_function(function) do
-    :lists.foldl(function, acc, list)
+  def foldl(list, acc, fun) when is_list(list) and is_function(fun) do
+    :lists.foldl(fun, acc, list)
   end
 
   @doc """
@@ -191,13 +245,13 @@ defmodule List do
 
   ## Examples
 
-      iex> List.foldr([1, 2, 3, 4], 0, fn(x, acc) -> x - acc end)
+      iex> List.foldr([1, 2, 3, 4], 0, fn x, acc -> x - acc end)
       -2
 
   """
   @spec foldr([elem], acc, (elem, acc -> acc)) :: acc when elem: var, acc: var
-  def foldr(list, acc, function) when is_list(list) and is_function(function) do
-    :lists.foldr(function, acc, list)
+  def foldr(list, acc, fun) when is_list(list) and is_function(fun) do
+    :lists.foldr(fun, acc, list)
   end
 
   @doc """
@@ -215,8 +269,9 @@ defmodule List do
       1
 
   """
-  @spec first([elem]) :: nil | elem when elem: var
-  def first([]),      do: nil
+  @spec first([]) :: nil
+  @spec first([elem, ...]) :: elem when elem: var
+  def first([]), do: nil
   def first([head | _]), do: head
 
   @doc """
@@ -234,16 +289,19 @@ defmodule List do
       3
 
   """
-  @spec last([elem]) :: nil | elem when elem: var
+  @spec last([]) :: nil
+  @spec last([elem, ...]) :: elem when elem: var
   def last([]), do: nil
   def last([head]), do: head
   def last([_ | tail]), do: last(tail)
 
   @doc """
   Receives a list of tuples and returns the first tuple
-  where the item at `position` in the tuple matches the
+  where the element at `position` in the tuple matches the
   given `key`.
 
+  If no matching tuple is found, `default` is returned.
+
   ## Examples
 
       iex> List.keyfind([a: 1, b: 2], :a, 0)
@@ -263,7 +321,7 @@ defmodule List do
 
   @doc """
   Receives a list of tuples and returns `true` if there is
-  a tuple where the item at `position` in the tuple matches
+  a tuple where the element at `position` in the tuple matches
   the given `key`.
 
   ## Examples
@@ -284,14 +342,17 @@ defmodule List do
   end
 
   @doc """
-  Receives a list of tuples and replaces the item
-  identified by `key` at `position` if it exists.
+  Receives a list of tuples and if the identified element by `key` at `position`
+  exists, it is replaced with `new_tuple`.
 
   ## Examples
 
       iex> List.keyreplace([a: 1, b: 2], :a, 0, {:a, 3})
       [a: 3, b: 2]
 
+      iex> List.keyreplace([a: 1, b: 2], :a, 1, {:a, 3})
+      [a: 1, b: 2]
+
   """
   @spec keyreplace([tuple], any, non_neg_integer, tuple) :: [tuple]
   def keyreplace(list, key, position, new_tuple) do
@@ -299,7 +360,7 @@ defmodule List do
   end
 
   @doc """
-  Receives a list of tuples and sorts the items
+  Receives a list of tuples and sorts the elements
   at `position` of the tuples. The sort is stable.
 
   ## Examples
@@ -317,10 +378,10 @@ defmodule List do
   end
 
   @doc """
-  Receives a `list` of tuples and replaces the item
-  identified by `key` at `position`.
+  Receives a `list` of tuples and replaces the element
+  identified by `key` at `position` with `new_tuple`.
 
-  If the item does not exist, it is added to the end of the `list`.
+  If the element does not exist, it is added to the end of the `list`.
 
   ## Examples
 
@@ -338,7 +399,7 @@ defmodule List do
 
   @doc """
   Receives a `list` of tuples and deletes the first tuple
-  where the item at `position` matches the
+  where the element at `position` matches the
   given `key`. Returns the new list.
 
   ## Examples
@@ -380,16 +441,16 @@ defmodule List do
   @spec keytake([tuple], any, non_neg_integer) :: {tuple, [tuple]} | nil
   def keytake(list, key, position) do
     case :lists.keytake(key, position + 1, list) do
-      {:value, item, list} -> {item, list}
-      false                -> nil
+      {:value, element, list} -> {element, list}
+      false -> nil
     end
   end
 
   @doc """
-  Wraps the argument in a list.
+  Wraps `term` in a list if this is not list.
 
-  If the argument is already a list, returns the list.
-  If the argument is `nil`, returns an empty list.
+  If `term` is already a list, it returns the list.
+  If `term` is `nil`, it returns an empty list.
 
   ## Examples
 
@@ -403,7 +464,9 @@ defmodule List do
       []
 
   """
-  @spec wrap(list | any) :: list
+  @spec wrap(term) :: maybe_improper_list()
+  def wrap(term)
+
   def wrap(list) when is_list(list) do
     list
   end
@@ -432,10 +495,96 @@ defmodule List do
   """
   @spec zip([list]) :: [tuple]
   def zip([]), do: []
+
   def zip(list_of_lists) when is_list(list_of_lists) do
     do_zip(list_of_lists, [])
   end
 
+  @doc ~S"""
+  Checks if `list` is a charlist made only of printable ASCII characters.
+
+  Takes an optional `limit` as a second argument. `ascii_printable?/2` only
+  checks the printability of the list up to the `limit`.
+
+  A printable charlist in Elixir contains only the printable characters in the
+  standard seven-bit ASCII character encoding, which are characters ranging from
+  32 to 126 in decimal notation, plus the following control characters:
+
+    * `?\a` - Bell
+    * `?\b` - Backspace
+    * `?\t` - Horizontal tab
+    * `?\n` - Line feed
+    * `?\v` - Vertical tab
+    * `?\f` - Form feed
+    * `?\r` - Carriage return
+    * `?\e` - Escape
+
+  For more information read the [Character groups](https://en.wikipedia.org/wiki/ASCII#Character_groups)
+  section in the Wikipedia article of the [ASCII](https://en.wikipedia.org/wiki/ASCII) standard.
+
+  ## Examples
+
+      iex> List.ascii_printable?('abc')
+      true
+
+      iex> List.ascii_printable?('abc' ++ [0])
+      false
+
+      iex> List.ascii_printable?('abc' ++ [0], 2)
+      true
+
+  Improper lists are not printable, even if made only of ASCII characters:
+
+      iex> List.ascii_printable?('abc' ++ ?d)
+      false
+
+  """
+  @doc since: "1.6.0"
+  @spec ascii_printable?(list, 0) :: true
+  @spec ascii_printable?([], limit) :: true
+        when limit: :infinity | pos_integer
+  @spec ascii_printable?([...], limit) :: boolean
+        when limit: :infinity | pos_integer
+  def ascii_printable?(list, limit \\ :infinity)
+      when is_list(list) and (limit == :infinity or (is_integer(limit) and limit >= 0)) do
+    ascii_printable_guarded?(list, limit)
+  end
+
+  defp ascii_printable_guarded?(_, 0) do
+    true
+  end
+
+  defp ascii_printable_guarded?([char | rest], counter)
+       # 7..13 is the range '\a\b\t\n\v\f\r'. 32..126 are ASCII printables.
+       when is_integer(char) and
+              ((char >= 7 and char <= 13) or char == ?\e or (char >= 32 and char <= 126)) do
+    ascii_printable_guarded?(rest, decrement(counter))
+  end
+
+  defp ascii_printable_guarded?([], _counter), do: true
+  defp ascii_printable_guarded?(_, _counter), do: false
+
+  @compile {:inline, decrement: 1}
+  defp decrement(:infinity), do: :infinity
+  defp decrement(counter), do: counter - 1
+
+  @doc """
+  Returns `true` if `list` is an improper list. Otherwise returns `false`.
+
+  ## Examples
+
+      iex> List.improper?([1, 2 | 3])
+      true
+
+      iex> List.improper?([1, 2, 3])
+      false
+
+  """
+  @doc since: "1.8.0"
+  @spec improper?(maybe_improper_list) :: boolean
+  def improper?(list) when is_list(list) and length(list) >= 0, do: false
+  def improper?(list) when is_list(list), do: true
+
   @doc """
   Returns a list with `value` inserted at the specified `index`.
 
@@ -458,11 +607,19 @@ defmodule List do
 
   """
   @spec insert_at(list, integer, any) :: list
-  def insert_at(list, index, value) when is_integer(index) do
-    if index < 0 do
-      do_insert_at(list, length(list) + index + 1, value)
-    else
-      do_insert_at(list, index, value)
+  def insert_at(list, index, value) when is_list(list) and is_integer(index) do
+    case index do
+      -1 ->
+        list ++ [value]
+
+      _ when index < 0 ->
+        case length(list) + index + 1 do
+          index when index < 0 -> [value | list]
+          index -> do_insert_at(list, index, value)
+        end
+
+      _ ->
+        do_insert_at(list, index, value)
     end
   end
 
@@ -488,9 +645,12 @@ defmodule List do
 
   """
   @spec replace_at(list, integer, any) :: list
-  def replace_at(list, index, value) when is_integer(index) do
+  def replace_at(list, index, value) when is_list(list) and is_integer(index) do
     if index < 0 do
-      do_replace_at(list, length(list) + index, value)
+      case length(list) + index do
+        index when index < 0 -> list
+        index -> do_replace_at(list, index, value)
+      end
     else
       do_replace_at(list, index, value)
     end
@@ -518,9 +678,12 @@ defmodule List do
 
   """
   @spec update_at([elem], integer, (elem -> any)) :: list when elem: var
-  def update_at(list, index, fun) when is_function(fun, 1) and is_integer(index) do
+  def update_at(list, index, fun) when is_list(list) and is_function(fun) and is_integer(index) do
     if index < 0 do
-      do_update_at(list, length(list) + index, fun)
+      case length(list) + index do
+        index when index < 0 -> list
+        index -> do_update_at(list, index, fun)
+      end
     else
       do_update_at(list, index, fun)
     end
@@ -567,6 +730,7 @@ defmodule List do
       {3, [1, 2]}
 
   """
+  @doc since: "1.4.0"
   @spec pop_at(list, integer, any) :: {any, list}
   def pop_at(list, index, default \\ nil) when is_integer(index) do
     if index < 0 do
@@ -576,18 +740,51 @@ defmodule List do
     end
   end
 
+  @doc """
+  Returns `true` if `list` starts with the given `prefix` list; otherwise returns `false`.
+
+  If `prefix` is an empty list, it returns `true`.
+
+  ### Examples
+
+      iex> List.starts_with?([1, 2, 3], [1, 2])
+      true
+
+      iex> List.starts_with?([1, 2], [1, 2, 3])
+      false
+
+      iex> List.starts_with?([:alpha], [])
+      true
+
+      iex> List.starts_with?([], [:alpha])
+      false
+
+  """
+  @doc since: "1.5.0"
+  @spec starts_with?(nonempty_list, nonempty_list) :: boolean
+  @spec starts_with?(list, []) :: true
+  @spec starts_with?([], nonempty_list) :: false
+  def starts_with?(list, prefix)
+
+  def starts_with?([head | tail], [head | prefix_tail]), do: starts_with?(tail, prefix_tail)
+  def starts_with?(list, []) when is_list(list), do: true
+  def starts_with?(list, [_ | _]) when is_list(list), do: false
+
   @doc """
   Converts a charlist to an atom.
 
-  Currently Elixir does not support conversions from charlists
-  which contains Unicode codepoints greater than 0xFF.
+  Elixir supports conversions from charlists which contains any Unicode
+  code point.
 
   Inlined by the compiler.
 
   ## Examples
 
-      iex> List.to_atom('elixir')
-      :elixir
+      iex> List.to_atom('Elixir')
+      :Elixir
+
+      iex> List.to_atom('🌢 Elixir')
+      :"🌢 Elixir"
 
   """
   @spec to_atom(charlist) :: atom
@@ -599,8 +796,8 @@ defmodule List do
   Converts a charlist to an existing atom. Raises an `ArgumentError`
   if the atom does not exist.
 
-  Currently Elixir does not support conversions from charlists
-  which contains Unicode codepoints greater than 0xFF.
+  Elixir supports conversions from charlists which contains any Unicode
+  code point.
 
   Inlined by the compiler.
 
@@ -610,6 +807,10 @@ defmodule List do
       iex> List.to_existing_atom('my_atom')
       :my_atom
 
+      iex> _ = :"🌢 Elixir"
+      iex> List.to_existing_atom('🌢 Elixir')
+      :"🌢 Elixir"
+
       iex> List.to_existing_atom('this_atom_will_never_exist')
       ** (ArgumentError) argument error
 
@@ -656,6 +857,8 @@ defmodule List do
 
   Inlined by the compiler.
 
+  The base needs to be between `2` and `36`.
+
   ## Examples
 
       iex> List.to_integer('3FF', 16)
@@ -684,11 +887,18 @@ defmodule List do
   end
 
   @doc """
-  Converts a list of integers representing codepoints, lists or
+  Converts a list of integers representing code points, lists or
   strings into a string.
 
-  Notice that this function expects a list of integers representing
-  UTF-8 codepoints. If you have a list of bytes, you must instead use
+  To be converted to a string, a list must either be empty or only
+  contain the following elements:
+
+    * strings
+    * integers representing Unicode code points
+    * a list containing one of these three elements
+
+  Note that this function expects a list of integers representing
+  Unicode code points. If you have a list of bytes, you must instead use
   the [`:binary` module](http://www.erlang.org/doc/man/binary.html).
 
   ## Examples
@@ -699,8 +909,14 @@ defmodule List do
       iex> List.to_string([0x0061, "bc"])
       "abc"
 
+      iex> List.to_string([0x0064, "ee", ['p']])
+      "deep"
+
+      iex> List.to_string([])
+      ""
+
   """
-  @spec to_string(:unicode.charlist) :: String.t
+  @spec to_string(:unicode.charlist()) :: String.t()
   def to_string(list) when is_list(list) do
     try do
       :unicode.characters_to_binary(list)
@@ -709,18 +925,71 @@ defmodule List do
         raise ArgumentError, """
         cannot convert the given list to a string.
 
-        To be converted to a string, a list must contain only:
+        To be converted to a string, a list must either be empty or only
+        contain the following elements:
 
           * strings
-          * integers representing Unicode codepoints
+          * integers representing Unicode code points
+          * a list containing one of these three elements
+
+        Please check the given list or call inspect/1 to get the list representation, got:
+
+        #{inspect(list)}
+        """
+    else
+      result when is_binary(result) ->
+        result
+
+      {:error, encoded, rest} ->
+        raise UnicodeConversionError, encoded: encoded, rest: rest, kind: :invalid
+
+      {:incomplete, encoded, rest} ->
+        raise UnicodeConversionError, encoded: encoded, rest: rest, kind: :incomplete
+    end
+  end
+
+  @doc """
+  Converts a list of integers representing Unicode code points, lists or
+  strings into a charlist.
+
+  Note that this function expects a list of integers representing
+  Unicode code points. If you have a list of bytes, you must instead use
+  the [`:binary` module](http://www.erlang.org/doc/man/binary.html).
+
+  ## Examples
+
+      iex> List.to_charlist([0x00E6, 0x00DF])
+      'æß'
+
+      iex> List.to_charlist([0x0061, "bc"])
+      'abc'
+
+      iex> List.to_charlist([0x0064, "ee", ['p']])
+      'deep'
+
+  """
+  @doc since: "1.8.0"
+  @spec to_charlist(:unicode.charlist()) :: charlist()
+  def to_charlist(list) when is_list(list) do
+    try do
+      :unicode.characters_to_list(list)
+    rescue
+      ArgumentError ->
+        raise ArgumentError, """
+        cannot convert the given list to a charlist.
+
+        To be converted to a charlist, a list must contain only:
+
+          * strings
+          * integers representing Unicode code points
           * or a list containing one of these three elements
 
         Please check the given list or call inspect/1 to get the list representation, got:
 
-        #{inspect list}
+        #{inspect(list)}
         """
     else
-      result when is_binary(result) ->
+      result when is_list(result) ->
         result
 
       {:error, encoded, rest} ->
@@ -744,89 +1013,142 @@ defmodule List do
   corresponding key is `:del`), or left alone (if the corresponding key is
   `:eq`) in `list1` in order to be closer to `list2`.
 
+  See `myers_difference/3` if you want to handle nesting in the diff scripts.
+
   ## Examples
 
       iex> List.myers_difference([1, 4, 2, 3], [1, 2, 3, 4])
       [eq: [1], del: [4], eq: [2, 3], ins: [4]]
 
   """
-  @spec myers_difference(list, list) :: [{:eq | :ins | :del, list}] | nil
+  @doc since: "1.4.0"
+  @spec myers_difference(list, list) :: [{:eq | :ins | :del, list}]
   def myers_difference(list1, list2) when is_list(list1) and is_list(list2) do
-    path = {0, 0, list1, list2, []}
-    find_script(0, length(list1) + length(list2), [path])
+    myers_difference_with_diff_script(list1, list2, nil)
   end
 
-  defp find_script(envelope, max, _paths) when envelope > max do
-    nil
+  @doc """
+  Returns a keyword list that represents an *edit script* with nested diffs.
+
+  This is an extension of `myers_difference/2` where a `diff_script` function
+  can be given in case it is desired to compute nested differences. The function
+  may return a list with the inner edit script or `nil` in case there is no
+  such script. The returned inner edit script will be under the `:diff` key.
+
+  ## Examples
+
+      iex> List.myers_difference(["a", "db", "c"], ["a", "bc"], &String.myers_difference/2)
+      [eq: ["a"], diff: [del: "d", eq: "b", ins: "c"], del: ["c"]]
+
+  """
+  @doc since: "1.8.0"
+  @spec myers_difference(list, list, (term, term -> script | nil)) :: script
+        when script: [{:eq | :ins | :del | :diff, list}]
+  def myers_difference(list1, list2, diff_script)
+      when is_list(list1) and is_list(list2) and is_function(diff_script) do
+    myers_difference_with_diff_script(list1, list2, diff_script)
   end
 
-  defp find_script(envelope, max, paths) do
-    case each_diagonal(-envelope, envelope, paths, []) do
+  defp myers_difference_with_diff_script(list1, list2, diff_script) do
+    path = {0, list1, list2, []}
+    find_script(0, length(list1) + length(list2), [path], diff_script)
+  end
+
+  defp find_script(envelope, max, paths, diff_script) do
+    case each_diagonal(-envelope, envelope, paths, [], diff_script) do
       {:done, edits} -> compact_reverse(edits, [])
-      {:next, paths} -> find_script(envelope + 1, max, paths)
+      {:next, paths} -> find_script(envelope + 1, max, paths, diff_script)
     end
   end
 
   defp compact_reverse([], acc), do: acc
 
+  defp compact_reverse([{:diff, _} = fragment | rest], acc) do
+    compact_reverse(rest, [fragment | acc])
+  end
+
   defp compact_reverse([{kind, elem} | rest], [{kind, result} | acc]) do
     compact_reverse(rest, [{kind, [elem | result]} | acc])
   end
 
+  defp compact_reverse(rest, [{:eq, elem}, {:ins, elem}, {:eq, other} | acc]) do
+    compact_reverse(rest, [{:ins, elem}, {:eq, elem ++ other} | acc])
+  end
+
   defp compact_reverse([{kind, elem} | rest], acc) do
     compact_reverse(rest, [{kind, [elem]} | acc])
   end
 
-  defp each_diagonal(diag, limit, _paths, next_paths) when diag > limit do
-    {:next, Enum.reverse(next_paths)}
+  defp each_diagonal(diag, limit, _paths, next_paths, _diff_script) when diag > limit do
+    {:next, :lists.reverse(next_paths)}
   end
 
-  defp each_diagonal(diag, limit, paths, next_paths) do
-    {path, rest} = proceed_path(diag, limit, paths)
-    with {:cont, path} <- follow_snake(path) do
-      each_diagonal(diag + 2, limit, rest, [path | next_paths])
+  defp each_diagonal(diag, limit, paths, next_paths, diff_script) do
+    {path, rest} = proceed_path(diag, limit, paths, diff_script)
+
+    case follow_snake(path) do
+      {:cont, path} -> each_diagonal(diag + 2, limit, rest, [path | next_paths], diff_script)
+      {:done, edits} -> {:done, edits}
     end
   end
 
-  defp proceed_path(0, 0, [path]), do: {path, []}
+  defp proceed_path(0, 0, [path], _diff_script), do: {path, []}
 
-  defp proceed_path(diag, limit, [path | _] = paths) when diag == -limit do
-    {move_down(path), paths}
+  defp proceed_path(diag, limit, [path | _] = paths, diff_script) when diag == -limit do
+    {move_down(path, diff_script), paths}
   end
 
-  defp proceed_path(diag, limit, [path]) when diag == limit do
-    {move_right(path), []}
+  defp proceed_path(diag, limit, [path], diff_script) when diag == limit do
+    {move_right(path, diff_script), []}
   end
 
-  defp proceed_path(_diag, _limit, [path1, path2 | rest]) do
-    if elem(path1, 1) > elem(path2, 1) do
-      {move_right(path1), [path2 | rest]}
+  defp proceed_path(_diag, _limit, [path1, path2 | rest], diff_script) do
+    if elem(path1, 0) > elem(path2, 0) do
+      {move_right(path1, diff_script), [path2 | rest]}
     else
-      {move_down(path2), [path2 | rest]}
+      {move_down(path2, diff_script), [path2 | rest]}
     end
   end
 
-  defp move_right({x, y, list1, [elem | rest], edits}) do
-    {x + 1, y, list1, rest, [{:ins, elem} | edits]}
+  defp move_right({y, [elem1 | rest1] = list1, [elem2 | rest2], edits}, diff_script)
+       when diff_script != nil do
+    if diff = diff_script.(elem1, elem2) do
+      {y + 1, rest1, rest2, [{:diff, diff} | edits]}
+    else
+      {y, list1, rest2, [{:ins, elem2} | edits]}
+    end
   end
 
-  defp move_right({x, y, list1, [], edits}) do
-    {x + 1, y, list1, [], edits}
+  defp move_right({y, list1, [elem | rest], edits}, _diff_script) do
+    {y, list1, rest, [{:ins, elem} | edits]}
   end
 
-  defp move_down({x, y, [elem | rest], list2, edits}) do
-    {x, y + 1, rest, list2, [{:del, elem} | edits]}
+  defp move_right({y, list1, [], edits}, _diff_script) do
+    {y, list1, [], edits}
   end
 
-  defp move_down({x, y, [], list2, edits}) do
-    {x, y + 1, [], list2, edits}
+  defp move_down({y, [elem1 | rest1], [elem2 | rest2] = list2, edits}, diff_script)
+       when diff_script != nil do
+    if diff = diff_script.(elem1, elem2) do
+      {y + 1, rest1, rest2, [{:diff, diff} | edits]}
+    else
+      {y + 1, rest1, list2, [{:del, elem1} | edits]}
+    end
   end
 
-  defp follow_snake({x, y, [elem | rest1], [elem | rest2], edits}) do
-    follow_snake({x + 1, y + 1, rest1, rest2, [{:eq, elem} | edits]})
+  defp move_down({y, [elem | rest], list2, edits}, _diff_script) do
+    {y + 1, rest, list2, [{:del, elem} | edits]}
   end
 
-  defp follow_snake({_x, _y, [], [], edits}) do
+  defp move_down({y, [], list2, edits}, _diff_script) do
+    {y + 1, [], list2, edits}
+  end
+
+  defp follow_snake({y, [elem | rest1], [elem | rest2], edits}) do
+    follow_snake({y + 1, rest1, rest2, [{:eq, elem} | edits]})
+  end
+
+  defp follow_snake({_y, [], [], edits}) do
     {:done, edits}
   end
 
@@ -842,10 +1164,6 @@ defmodule List do
     []
   end
 
-  defp do_replace_at(list, index, _value) when index < 0 do
-    list
-  end
-
   defp do_replace_at([_old | rest], 0, value) do
     [value | rest]
   end
@@ -860,7 +1178,7 @@ defmodule List do
     [value]
   end
 
-  defp do_insert_at(list, index, value) when index <= 0 do
+  defp do_insert_at(list, 0, value) do
     [value | list]
   end
 
@@ -874,10 +1192,6 @@ defmodule List do
     [fun.(value) | list]
   end
 
-  defp do_update_at(list, index, _fun) when index < 0 do
-    list
-  end
-
   defp do_update_at([head | tail], index, fun) do
     [head | do_update_at(tail, index - 1, fun)]
   end
@@ -892,10 +1206,6 @@ defmodule List do
     {default, :lists.reverse(acc)}
   end
 
-  defp do_pop_at(list, index, default, []) when index < 0 do
-    {default, list}
-  end
-
   defp do_pop_at([head | tail], 0, _default, acc) do
     {head, :lists.reverse(acc, tail)}
   end
@@ -908,8 +1218,11 @@ defmodule List do
 
   defp do_zip(list, acc) do
     converter = fn x, acc -> do_zip_each(to_list(x), acc) end
+
     case :lists.mapfoldl(converter, [], list) do
-      {_, nil} -> :lists.reverse(acc)
+      {_, nil} ->
+        :lists.reverse(acc)
+
       {mlist, heads} ->
         do_zip(mlist, [to_tuple(:lists.reverse(heads)) | acc])
     end
@@ -928,5 +1241,5 @@ defmodule List do
   end
 
   defp to_list(tuple) when is_tuple(tuple), do: Tuple.to_list(tuple)
-  defp to_list(list)  when is_list(list),   do: list
+  defp to_list(list) when is_list(list), do: list
 end
